

                        /*   MS3  */

               /* Adapted for MPE display  */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>
#include <mpe.h>
#include <mpe_graphics.h>


                /********************************/
                /*                              */
                /*     The Mandelbrot Set       */
                /*                              */
                /*      z = z*z + c             */
                /*                              */
                /********************************/

/*  Simplified main() via two functions   */
/*  iterate_map() and set_grid()          */

/*  Slice image - strips parallel to real C */
/*  via functions                           */
/*  set_slices() and calculate_slice()      */

/*  Parallel version -- Master-Slave model  */
/*                                          */
/*  master()  running on node zero          */
/*  slave()   running on all other nodes    */
/*                                          */


/* use:                  */
/*                       */
/* MS3                   */


                                        /* Michel Vallieres */


                                        /* large distance; if |z| goes  */
                                        /* beyond it, point C is not a  */
                                        /* member of the set            */
#define LARGE_RADIUS 2.0
                                        /* Maximum number of iterations */
                                        /* before declaring a point in  */
                                        /* the Mandelbrot set           */
#define MAX_ITERATIONS 120
                                        /* image size */
#define NX 700
#define NY 500
                                        /* # of slices parallel to C real */
#define N_slices 64
                                        /* Maximum # processes */
#define MAX_PROCESSES 1000
                                        /* # of colors in MPE palette */
#define NC 128



/* Iterate the Mandelbrot map -- decide if the (input) point  */
/* in complex C plane belongs or not to the set               */
/* prints out the number of iterations pro-rated to           */
/* MAX_ITERATIONS                                             */
double iterate_map( double c_real, double c_img, int *count )
{
   double z_real, z_img, z_current_real, z_magnitude;
   double color;
   int    counter;

                                        /* z starts a origin */
   z_real = 0.0;
   z_img  = 0.0;
                                        /* set counter to zero */
   counter = 0;
                                        /* iterate map for             */
                                        /*  MAX_ITERATIONS times or ...*/
   while ( counter < MAX_ITERATIONS ) 
   {
     z_current_real = z_real;
     z_real = z_real*z_real - z_img*z_img + c_real;
     z_img  = 2.0*z_current_real*z_img + c_img;
     counter = counter + 1;
                                        /* ... until magnitude of z */
                                        /* is larger than some      */
                                        /* large radius             */
     z_magnitude = z_real*z_real+z_img*z_img;
     if ( z_magnitude > LARGE_RADIUS ) break;
   }
                                        /* write C real and       */
                                        /* imaginary parts if     */
                                        /* C belongs to the set   */
         /*
           if ( z_magnitude < LARGE_RADIUS ) 
           printf ( " %f %f  \n ", c_real, c_img );
	 */
	                                /* or iteration number      */
	                                /* for all scanned C values */
	                                /* for color picture        */
   count++;

   color = (double)(counter) / (double)MAX_ITERATIONS;
   return color;

}


/* Set up the grid in complex C plane to generate image  */
void set_grid( double *crmin, double *crmax, double *cimin,
               double *cimax, double *dcr, double *dci )
{
                                        /* define grid spanning   */
                                        /* the complex plane for  */
                                        /* the C variable         */
   *crmin = -1.7;
   *crmax = 0.8;
   *cimin = -1.1;
   *cimax = 1.1;
                                        /* C spacing on complex plane */
   *dcr = (*crmax - *crmin)/(NX-1);
   *dci = (*cimax - *cimin)/(NY-1);
}


/* Calculate the Mandelbrot set in slice  */
void calculate_slice( int slice, int *begin_slice,
            int *end_slice, int *count,
            double crmin, double cimin, double dcr, double dci,
		      double *store_slice, MPE_Point store_points[] )
{
  int    i, j, k;
  double c_real, c_img, color;
  int    color_pointer;
                                        /* store pointer */
  k = 0;
                                        /* scan over slice */
  for ( j=begin_slice[slice] ; j<=end_slice[slice] ; j++ )
  {
    for ( i=0 ; i<NX ; i++ )
    {
       c_real = crmin + i*dcr;
       c_img = cimin + j*dci;
                                        /* iterate Mandelbrot map */
       color = iterate_map( c_real, c_img, count );

                                        /* store iteration # for */
                                        /* use in MPE display    */
                                        /* color is NC scaled by */
                                        /* iterations/max_iterations */
       store_points[k].x = i;
       store_points[k].y = j;
       color_pointer = NC*color;
       store_points[k].c = color_pointer;

                                        /* store resulting color */
       store_slice[k] = color;
       k++;
    }
  }
}


/* Setup the slices through which compute the Mandelbrot set  */
int set_slices( int *begin_slice, int *end_slice, int *offset )
{
  int slice, average, leftover, temp, temp_offset;
  MPI_Status status;

  average = NY / N_slices;
  leftover = NY % N_slices;
  temp = -1;
  temp_offset = 0;
  for ( slice=0 ; slice<N_slices ; slice++ )
    {
      begin_slice[slice] = temp + 1;
      end_slice[slice] = begin_slice[slice] + average - 1;
      if ( leftover >0 )
        {
          end_slice[slice]++;
          leftover--;
        }
      temp = end_slice[slice];
      offset[slice] = temp_offset;
      temp_offset = temp_offset + NX * 
              ( end_slice[slice] - begin_slice[slice] + 1 );
    }
  return average;
}


/* terminates graph interactively */
int quit_by_touch( MPE_XGraph graph, int WINDOW_SIZE_X, 
                   int WINDOW_SIZE_Y  )
{
  int x, y, button, xc, yc, w, h, xs, ys;

  xc = 0.95*WINDOW_SIZE_X;              /* draw filled rectangle     */
  yc = 0.95*WINDOW_SIZE_Y;              /* lower left bottom         */
  w = 0.05*WINDOW_SIZE_X;               /* corner of graphic window  */
  h = WINDOW_SIZE_Y - yc;
                                        /* write Q in small rectangle */
  MPE_Fill_rectangle( graph, xc, yc, w, h, MPE_BLACK );
  xs = xc + 0.4*w;
  ys = yc + 0.8*h;
  MPE_Draw_string( graph, xs, ys, MPE_WHITE, "Q" );

                                        /* infinite loop            */
                                        /* wait for different touch */
  for ( ; ; )
    {
	                                /* wait for mouse to touch graph */
                                        /* can only be done by master    */
      MPE_Get_mouse_press( graph, &x, &y, &button );
	                                /* result is touch point */
      fprintf( stderr, "\n Graph coordinates of the touch: %d %d %d \n", 
	               x, y, button );
      if ( x < xc+w && y > yc )  
	{
	  fprintf( stderr, 
                   "\n Last touch was in the -- quit -- rectangle \n\n" );
	  return;
	}  
    }
}



/* slave processes -- calculate a specific slice   */
/* specified by "slice" as received from process 0 */
void slave( int my_rank, int *begin_slice, int *end_slice,
            int average, double crmin, double cimin, double dcr, 
            double dci, double *store_slice , MPE_Point store_points[],
            MPE_XGraph graph )
{
  static int count;
  int        number, store_n, slice;
  int        done;
  MPI_Status status;
                                         /* signal ->  slice is done */
  done = 1;

  for ( ; ; )
    {
                                        /* a new slice to calculate */
      MPI_Recv( &slice, 1, MPI_INT, 0, 325, MPI_COMM_WORLD, &status );

                                        /* suicide signal */
      if ( slice < 0 )
	{
          free( store_slice );
	  free( store_points );
          MPE_Close_graphics( &graph );
	  MPI_Finalize();
	  exit(0);
	}
                                       /* calculate requested slice */
      calculate_slice( slice, begin_slice, end_slice, &count,
		       crmin, cimin, dcr, dci, store_slice,
                       store_points );

                                       /* send results back to master */
      //  number = NX * (average+1) ;
      //  MPI_Ssend( store_slice, number, MPI_DOUBLE, 0, 327, 
      //  	        MPI_COMM_WORLD );

                                       /* send signal to node 0 */
      MPI_Ssend( &done, 1, MPI_INT, 0, 327, MPI_COMM_WORLD );
      
                                       /* draw the slice in MPE graph */
      store_n = NX * ( end_slice[slice]-begin_slice[slice]+1 );
      MPE_Draw_points( graph, store_points, store_n);

                                        /* update the graph */
      MPE_Update( graph );

      fprintf( stderr, "slave %d  ->  slice %d calculated & drawn\n", 
                            my_rank, slice );
    }
}



/* Master node -- control the calculation */
void master(int numprocs, int *begin_slice, int *end_slice, 
            int average, double crmin, double cimin, double dcr, 
            double dci, double *store_slice, int *offset,
            MPE_XGraph graph )
{
  int    k, k_slice, slice, process;
  int    number, source;
  int    done;
  static int count;
  static int list_slices[MAX_PROCESSES];
  int    kill, recv_slice;
  double mandelbrot[NX*NY];
  MPI_Status status;
                                        /* # points in set */
   count = 0;
                                        /* scan over slices */
   slice = 0;
                                        /* send a slice to each slave */
   for ( process=1 ; process<numprocs ; process++ )
     { 
                                        /* generate Mandelbrot set */
                                        /*  in each process        */
       MPI_Ssend( &slice, 1, MPI_INT, process, 325, 
		MPI_COMM_WORLD );
       list_slices[process] = slice;
       slice++;
     }

                                        /* scan over received slices */
   for ( k_slice=0 ; k_slice<N_slices ; k_slice++ )
     {
                                        /* receive a slice */
       // number = NX * ( average+1);
       // MPI_Recv( store_slice, number, MPI_DOUBLE, MPI_ANY_SOURCE, 327, 
       //	        MPI_COMM_WORLD, &status );

                                        /* watch for done signal */
       MPI_Recv( &done, 1, MPI_INT, MPI_ANY_SOURCE, 327, 
		        MPI_COMM_WORLD, &status );

                                        /* source of slice */
       source = status.MPI_SOURCE;
                                        /* received slice */ 
       recv_slice = list_slices[source];
                                        /* send a slice to this slave */
       if ( slice < N_slices )
	 {      
            MPI_Ssend( &slice, 1, MPI_INT, source, 325, 
                                        MPI_COMM_WORLD );
            list_slices[source] = slice;
            slice++;
	 }
                                        /* actual number */
       // number = NX * ( end_slice[recv_slice]-begin_slice[recv_slice]+1 );

                                        /* store set in matrix */
       // for ( k=0 ; k<number ; k++ )
       //   mandelbrot[offset[recv_slice]+k] = store_slice[k];

     }   

                                        /* ouput Mandelbrot set */
   // fprintf( stderr, " The Mandelbrot Set will be written out \n" );
   // for ( k=0; k<NX*NY ; k++ )
   //        printf( " %f \n", mandelbrot[k] );

   // fprintf( stderr, " The Mandelbrot Set has been written out \n" );

                                        /* calculation is done!    */
                                        /* all stripes done        */
                                        /* wait for a touch - quit */
   sleep( 2 );
   quit_by_touch( graph, NX, NY );

                                        /* end all slave processes */
   kill = -1;
   for ( process=1 ; process<numprocs ; process++ )
     { 
       MPI_Ssend( &kill, 1, MPI_INT, process, 325, MPI_COMM_WORLD );
     }
                                        /* end itself */
   free( store_slice );
   MPE_Close_graphics( &graph );
   MPI_Finalize();
   exit( 0 );
}



int main( int argc, char *argv[] )
{
   double crmin, crmax, cimin, cimax; 
   double dcr, dci;
   int    average;
   int    begin_slice[1000], end_slice[1000], offset[1000];
   int    my_rank, numprocs;
   double *store_slice;
   MPE_Point *store_points;
   MPE_Color colors[500];
   MPE_XGraph graph;
   int    ierr;

   char displayname[MPI_MAX_PROCESSOR_NAME+4] = "";

                            /* MPI exposure  */
   MPI_Init( &argc, &argv );
   MPI_Comm_size( MPI_COMM_WORLD, &numprocs );
   MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

   if ( my_rank == 0 )
        strcpy( displayname, getenv( "DISPLAY" ) );

   MPI_Bcast( displayname, MPI_MAX_PROCESSOR_NAME+4, MPI_CHAR, 
               0, MPI_COMM_WORLD );
   fprintf( stdout, "%d : $DISPLAY at process 0 = %s\n",
             my_rank, displayname );
   fflush( stdout );

   ierr = MPE_Open_graphics( &graph, MPI_COMM_WORLD, displayname,
                              -1, -1, 700, 500, 0 );

   if ( ierr != MPE_SUCCESS ) 
     {
       fprintf( stderr, "%d : MPE_Open_graphics() fails\n", my_rank );
       ierr = MPI_Abort( MPI_COMM_WORLD, 1 );
       exit(1);
     }

   fprintf( stderr, "After Open Graphics call \n" );

                                        /* make a new color palette */
   MPE_Make_color_array( graph, NC, colors );

                                        /* set slicing */
   average = set_slices( begin_slice, end_slice, offset );

                                        /* setup C grid */
   set_grid( &crmin, &crmax, &cimin, &cimax, &dcr, &dci );
 
                                        /* color storage */
   store_slice = (double *)malloc( NX*(average+1)*sizeof(double) );

                                        /* color storage for MPE graph */
   store_points = (MPE_Point *)malloc( NX*(average+1)*sizeof(MPE_Point) );

                                        /* master node */
   if ( my_rank == 0 )
     {
       master( numprocs, begin_slice, end_slice, average,
	       crmin, cimin, dcr, dci, store_slice, offset, graph );
     }
                                        /* slave nodes  */
   else
     {
       slave( my_rank, begin_slice, end_slice, average,
	      crmin, cimin, dcr, dci, store_slice, store_points, graph );
     }

}


